"""
The RDFToLiPD class helps in converting an RDF Graph to a LiPD file.
It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion
"""

import copy
import os
import re
import csv
import bagit
import json
import tempfile
import zipfile

from rdflib.graph import URIRef

from ..globals.urls import NSURL
from ..globals.blacklist import REVERSE_BLACKLIST
from ..globals.schema import SCHEMA
from ..globals.synonyms import RSYNONYMS

from .utils import ucfirst, lcfirst, unzip_string

class RDFToLiPD:
    """
    The RDFToLiPD class helps in converting an RDF Graph to a LiPD file.
    It uses the SCHEMA dictionary (from globals/schema.py) to do the conversion
    """    
    def __init__(self, graph):
        self.graph = graph
        self.lipd_csvs = {}
        self.graphurl = NSURL
        self.namespace = NSURL + "/"

    def convert(self, dsname, lipdfile):
        '''Convert RDF graph to a LiPD file

        Parameters
        ----------

        graph : rdflib.ConjunctiveGraph
            the RDF graph object

        '''
        lipd = self.convert_to_json(dsname)

        with tempfile.TemporaryDirectory(prefix="rdf_to_lipd_") as tmpdir:
            # Create a temporary data directory
            datadir = f"{tmpdir}/{dsname}"
            os.makedirs(datadir)

            # Create the csv files and metadata jsonld
            self._create_csvs(lipd, datadir)
            with(open(f"{datadir}/metadata.jsonld", "w")) as f:
                json.dump(lipd, f, indent=4, default=str)
            
            # Convert data directory to a bag
            bagit.make_bag(datadir, checksums=['md5'])

            # Zip the bag
            self._zip_directory(datadir, lipdfile)

        return lipd

    def convert_to_json(self, dsname):
        '''Convert RDF graph to a LiPD file

        Parameters
        ----------

        graph : rdflib.ConjunctiveGraph
            the RDF graph object

        '''
        self.schema = copy.deepcopy(SCHEMA)
        self.rschema = self._get_schema_reverse_map()

        self.allfacts = {}
        self._get_indexed_facts(self.namespace + dsname)
        
        lipd = self._convert_to_lipd(self.namespace + dsname, "Dataset", "Dataset", pagesdone={})
        lipd = self._post_processing(lipd)
        return lipd

    def _get_table_data(self, table):
        csvdata = []
        numrows = 0
        
        table["columns"] = sorted(table["columns"], key=lambda x: x["number"] if ("number" in x and type(x["number"]) is int) else 99999)

        for col in table["columns"]:
            if "values" in col:
                numrows = len(col["values"])
                break
        
        for rowidx in range(numrows):
            row = []
            for col in table["columns"]:
                if "values" in col:
                    colindices = []
                    if "number" in col:
                        colnum = col["number"]
                        colindices.append(colnum)
                    else:
                        colindices.append(1)
                    
                    for colindex in colindices:
                        if type(colindex) is list:
                            sorted(colindex)
                            firstidx = colindex[0]
                            for subindex in colindex:
                                idx = subindex - 1
                                if idx >= len(row):
                                    row.extend([""] * (idx - len(row) + 1))
                                row[idx] = col["values"][rowidx][subindex-firstidx]

                        else:
                            idx = colindex - 1
                            if idx >= len(row):
                                row.extend([""] * (idx - len(row) + 1))
                            row[idx] = col["values"][rowidx]

            csvdata.append(row)

        # Delete the values field (since we've extracted the csv data)
        for col in table["columns"]:
            if "values" in col:
                del col["values"]

        return csvdata
    
    def _create_csvs(self, lipd, datadir):
        csvs = {}
        datakeys = ["paleoData", "chronData"]
        for datakey in datakeys:
            if datakey in lipd:
                for data in lipd[datakey]:
                    if "measurementTable" in data:
                        for table in data["measurementTable"]:
                            csvs[table["filename"]] = self._get_table_data(table)
                    if "model" in data:
                        for model in data["model"]:
                            if "ensembleTable" in model:
                                for table in model["ensembleTable"]:
                                    csvs[table["filename"]] = self._get_table_data(table)
                            if "summaryTable" in model:
                                for table in model["summaryTable"]:
                                    csvs[table["filename"]] = self._get_table_data(table)

        for csvname, csvdata in csvs.items():
            # writing to csv file 
            with open(f"{datadir}/{csvname}", 'w') as csvfile: 
                csvwriter = csv.writer(csvfile) 
                csvwriter.writerows(csvdata)


    def _zip_directory(self, datadir, lipdfile):
        # Zip the bag
        with zipfile.ZipFile(lipdfile, 'w') as zipf:
            for root, dirs, files in os.walk(datadir):
                for file in files:
                    zipf.write(os.path.join(root, file), os.path.relpath(os.path.join(root, file), os.path.join(datadir, '..')))

    '''
    def _zip_lipd_dir(self, lipddir, lipdfile):
        if lipdfile.startswith("http"):
            # If this is a URL
            # Handle special characters in url (if any)
            res = urlparse(lipdfile)
            lipdurl = urlunparse(res._replace(path=quote(res.path)))

            # Open url and unzip
            resp = urlopen(lipdurl)
            with zipfile.ZipFile(BytesIO(resp.read())) as zip_ref:
                zip_ref.extractall(unzipdir)
        else:
            # If this is a local file
            # Unzip file
            with zipfile.ZipFile(lipdfile, 'r') as zip_ref:
                zip_ref.extractall(unzipdir)
    '''
    
    def _get_property_details(self, pname, schema) :
        details = { "name": pname }
        # Create property
        if schema and (pname in schema) :
            for skey,svalue in schema[pname].items() :
                details[skey] = svalue
        return details

    def _get_rdf_property_details(self, pname, fullkey, schema) :
        key = pname
        pname = lcfirst(pname)
        details = { "name": pname }
		# Check for full key in schema
        if schema and (fullkey in schema):
            for ind, svalue in schema[fullkey].items():
                details[ind] = svalue
		
        # Or check for just pname in schema
        elif schema and (key in schema):
            for ind, svalue in schema[key].items():
                details[ind] = svalue

        return details;


    def _get_schema_reverse_map(self) :
        newschema = {}
        for schid,sch in self.schema.items() :
            newsch = {}
            for prop,details in sch.items() :
                if (prop[0] == "@") :
                    continue
                
                if (("skip_auto_convert_to_json" in details)) :
                    continue
                
                pdetails = self._get_property_details(prop, sch)
                pname = pdetails["name"]
                pdetails["name"] = prop
                newsch[pname] = pdetails
                if (("category" in pdetails)) :
                    catpname = pname + "." + ucfirst(pdetails["category"])
                    newsch[catpname] = pdetails
                
                if (("schema" in pdetails)) :
                    schpname = pname + "." + ucfirst(pdetails["schema"])
                    newsch[schpname] = pdetails

            newschema[schid] = newsch
        
        return newschema

    def _local_name(self, url):
        return str(url).split("#")[-1]

    def _get_prop_values_from_query_result_p_o(self, qres):
        result = {}
        for row in qres:
            pname = self._local_name(row.p)
            if pname not in result:
                result[pname] = []
            value = {}
            if isinstance(row.o, URIRef):
                value["@type"] = "uri"
                value["@id"] = str(row.o)
            else:
                value["@type"] = "literal"
                value["@value"] = row.o.value
                value["@datatype"] = str(row.o.datatype)
            result[pname].append(value)
        return result

    def _get_facts(self, id):
        query = f"SELECT ?p ?o WHERE {{ <{id}> ?p ?o }}"
        qres = self.graph.query(query)
        return self._get_prop_values_from_query_result_p_o(qres)

    def _get_indexed_facts(self, id):
        if id in self.allfacts:
            return

        facts = self._get_facts(id)
        self.allfacts[id] = facts

        for pname, pfacts in facts.items():
            for pfact in pfacts:
                if pfact["@type"] == "uri":
                    if pname != "type":
                        self._get_indexed_facts(pfact["@id"])


    def _convert_to_lipd(self, id, category, schemaname, pagesdone={}) :
        if id in self.allfacts:
            facts = self.allfacts[id]

            if id in pagesdone:
                return pagesdone[id]
            
            schema =  self.rschema[schemaname] if (schemaname in self.rschema) else None
            if (schemaname and not category) :
                category = schemaname
            
            if "type" in facts:
                cats = facts["type"]
                for cat in cats:
                    if (cat["@type"] == "uri") :
                        category = self._local_name(cat["@id"])
                        break
            
            obj = {
                "@id":id,
                "@category":category,
                "@schema":schemaname
            }

            pagesdone[id] = obj

            for pname, pfacts in facts.items() :
                if pname in REVERSE_BLACKLIST:
                    continue
                prop = pname
                prop = re.sub("/\\s/", "_", prop)
                # Get a sample value page category, and use to make a property key
                propkey = prop
                for value in pfacts: 
                    if value["@type"] == "uri" :
                        if value["@id"] in self.allfacts :
                            pfact = self.allfacts[value["@id"]]
                            if "type" in pfact:
                                valcats = pfact["type"]
                                for valcat in valcats:
                                    if (valcat["@type"] == "uri") :
                                        valcatname = self._local_name(valcat["@id"])
                                        propkey = prop + "." + valcatname
                                        break

                details = self._get_rdf_property_details(prop, propkey, schema)
                name = details["name"]
                ptype =  details["type"] if ("type" in details) else None
                cat =  details["category"] if ("category" in details) else None
                sch =  details["schema"] if ("schema" in details) else None
                if (cat and not sch) :
                    sch = cat
                
                toJson =  details["toJson"] if ("toJson" in details) else None
                multiple =  details["multiple"] if ("multiple" in details) else False

                if len(pfacts) > 0 :
                    if (multiple) :
                        obj[name] = []
                    
                    for pfact in pfacts : 
                        if pfact["@type"] == "uri" :
                            val = self._convert_to_lipd(pfact["@id"], cat, sch, pagesdone)
                        else:
                            val = pfact["@value"]
                        
                        if (toJson) :
                            fn = getattr(self, toJson)
                            val = fn(val)
                                                
                        # If there is already a value present
                        # - Then this need to be marked as "multiple"
                        if (not multiple) and (name in obj) and (not type(obj[name]) is list):
                            multiple = True
                            obj[name] = [obj[name]]
                        
                        if (multiple) :
                            obj[name].append(val)
                        else : 
                            obj[name] = val

            return obj
        else : 
            return re.sub("/_/", " ", id)


    def _post_processing(self, obj, parent=None):
        if not(type(obj) is dict):
            return obj
		
        if not("@schema" in obj):
            return obj

        # FIXME: Hack to avoid recursion
        schemaname = obj["@schema"]
        tschema = self.schema[schemaname] if schemaname in self.schema else None

        if tschema and "@toJson_pre" in tschema:
            for func in tschema["@toJson_pre"]:
                fn = getattr(self, func)
                obj = fn(obj, parent)        

        for key,value in obj.items():
            if type(value) is list:
                for i in range(len(value)):
                    obj[key][i] = self._post_processing(value[i], obj)
            else:
                obj[key] = self._post_processing(value, obj)

        if tschema and "@toJson" in tschema:
            for func in tschema["@toJson"]:
                fn = getattr(self, func)
                obj = fn(obj, parent)

        # FIXME: Only remove hasValues.. LiPD CSVs should be created outside this function beforehand
        if "hasValues" in obj:
            valuestr = obj["hasValues"]
            obj["values"] = json.loads(valuestr)
            del obj["hasValues"]
        
        del obj["@id"]
        del obj["@schema"]
        del obj["@category"]
        if "type" in obj:
            del obj["type"] 

        return obj

    #########################################
    # Converters
    #########################################
    def _location_to_json(self, geo, parent = None) :
        geojson = {
            "geometry":
                {
                    "coordinates" : [0, 0, 0],
                },
            "properties": {}
        }
        if "coordinates" in geo :
            latlong = geo["coordinates"].split(",")
            geojson["geometry"]["coordinates"] = [ float(latlong[1]), float(latlong[0]), float(latlong[2]) if len(latlong)>2 else 0 ]
            geojson["geometry"]["type"] = "Point"
        
        if "long" in geo :
            geojson["geometry"]["coordinates"][0] = float(geo["long"])
        if "longitude" in geo :
            geojson["geometry"]["coordinates"][0] = float(geo["longitude"])
        
        if "lat" in geo :
            geojson["geometry"]["coordinates"][1] = float(geo["lat"])
        if "latitude" in geo :
            geojson["geometry"]["coordinates"][1] = float(geo["latitude"])
        
        if "alt" in geo and geo["alt"] != "NA":
            geojson["geometry"]["coordinates"][2] = float(geo["alt"])
        if "elevation" in geo and geo["elevation"] != "NA":
            geojson["geometry"]["coordinates"][2] = float(geo["elevation"])

        for prop,value in geo.items() :
            if prop[0] == "@" :
                continue
            
            if prop == "locationType" :
                geojson["type"] = geo["locationType"]
            else : 
                if prop == "coordinates" or prop == "coordinatesFor":
                    # Ignore
                    pass
                else : 
                    if re.search(r"^(geo|wgs84):", prop) :
                        # Ignore
                        pass
                    elif prop in ["long", "lat", "alt"] :
                        pass
                    else : 
                        geojson["properties"][prop] = value
        
        return geojson


    def _get_google_spreadsheet_key(self, url:str, parent = None) :
        return url.replace("https://docs.google.com/spreadsheets/d/", "")

    def _remove_found_in_table(self, var, parent = None) :
        if "foundInTable" in var :
            del var["foundInTable"]
        return var

    def _remove_found_in_dataset(self, var, parent = None) :
        if "foundInDataset" in var :
            del var["foundInDataset"]
        return var

    def _unwrap_uncertainty(self, var, parent = None) :
        if (("hasUncertainty" in var)) :
            unc = var["hasUncertainty"]
            if (("hasValue" in unc)) :
                var["uncertainty"] = float(unc["hasValue"])
                del unc["hasValue"]
            
            for key,value in unc.items() :
                if (key[0] != "@") :
                    var[key] = value
                
            del var["hasUncertainty"]
        return var

    def _unwrap_integration_time(self, interp, parent = None) :
        if (("integrationTime" in interp)) :
            intime = interp["integrationTime"]
            if (("hasValue" in intime)) :
                interp["integrationTime"] = float(intime["hasValue"])
                del intime["hasValue"]
            
            for key,value in intime.items() :
                if (key[0] != "@") :
                    interp["integrationTime" + ucfirst(key)] = value

            del interp["hasIntegrationTime"]
        return interp

    def _collect_variables_by_id(self, item, arr) :
        if not (type(item) is dict):
            return arr

        # Data is a Hash
        if ("@category" in item) and ("@id" in item) and re.match("/Variable\$/", item["@category"]) :
            arr[item["@id"]] = item
        else : 
            for key,value in item.items() :
                if (key[0] != "@") :
                    arr = self._collect_variables_by_id(item[key], arr)
        return arr


    def _set_archive_type_label(self, ds, parent = None) :
        if "hasArchiveType" in ds :
            if "@id" in ds["hasArchiveType"]:
                id = ds["hasArchiveType"]["@id"]
                if id in RSYNONYMS:
                    ds["archiveType"] = RSYNONYMS[id]
                else:
                    ds["archiveType"] = ds["hasArchiveType"]["label"]
            del ds["hasArchiveType"]
        return ds


    def _set_variable_name_from_standard_variable_label(self, var, parent = None) :
        if "hasStandardVariable" in var :
            if "@id" in var["hasStandardVariable"]:
                id = var["hasStandardVariable"]["@id"]
                if id in RSYNONYMS:
                    var["variableName"] = RSYNONYMS[id]
                else:
                    var["variableName"] = var["hasStandardVariable"]["label"]
            del var["hasStandardVariable"]
        return var
    

    def _set_units_label(self, var, parent = None) :
        if "hasUnits" in var :
            if "@id" in var["hasUnits"]:
                id = var["hasUnits"]["@id"]
                if id in RSYNONYMS:
                    var["units"] = RSYNONYMS[id]
                else:
                    var["units"] = var["hasUnits"]["label"]
            del var["hasUnits"]
        return var
    

    def _set_proxy_label(self, var, parent = None) :
        if "hasProxy" in var :
            if "@id" in var["hasProxy"]:
                id = var["hasProxy"]["@id"]
                if id in RSYNONYMS:
                    var["proxy"] = RSYNONYMS[id]
                else:
                    var["proxy"] = var["hasProxy"]["label"]
            del var["hasProxy"]
        return var
    

    def _set_proxy_general_label(self, var, parent = None) :
        if "hasProxyGeneral" in var :
            if "@id" in var["hasProxyGeneral"]:
                id = var["hasProxyGeneral"]["@id"]
                if id in RSYNONYMS:
                    var["proxyGeneral"] = RSYNONYMS[id]
                else:
                    var["proxyGeneral"] = var["hasProxyGeneral"]["label"]
            del var["hasProxyGeneral"]
        return var
    
    def _set_interpretation_variable_label(self, interp, parent = None) :
        if "hasVariable" in interp :
            if "@id" in interp["hasVariable"]:
                id = interp["hasVariable"]["@id"]
                if id in RSYNONYMS:
                    interp["variable"] = RSYNONYMS[id]
                else:
                    interp["variable"] = interp["hasVariable"]["label"]
            del interp["hasVariable"]
        return interp
    
    def _set_seasonality_labels(self, interp, parent = None) :
        convs = {
            "hasSeasonality": "seasonality",
            "hasSeasonalityGeneral": "seasonalityGeneral",
            "hasSeasonalityOriginal": "seasonalityOriginal"
        }
        for pid in convs:
            if pid in interp :
                if "@id" in interp[pid]:
                    id = interp[pid]["@id"]
                    nid = convs[pid]
                    if id in RSYNONYMS:
                        interp[nid] = RSYNONYMS[id]
                    else:
                        interp[nid] = interp[pid]["label"]
                del interp[pid]
        return interp


    def _create_publication_identifier(self, pub, parent = None) :
        identifiers = []
        if (("hasDOI" in pub)) :
            identifier = {}
            identifier["type"] = "doi"
            identifier["id"] = pub["hasDOI"]
            if (("link" in pub)) :
                for link in pub["link"].values() : 
                    if (re.match("/dx.doi.org/", link)) :
                        identifier["url"] = link
                del pub["link"]
            del pub["hasDOI"]
            identifiers.append(identifier)

        pub["identifier"] = identifiers
        return pub


    def _values_to_array(self, resolution, parent = None) :
        if (("values" in resolution)) :
            return resolution["values"].split(",")

    def _unarray_column_number(self, var, parent = None) :
        if not var:
            return var
        if ("number" in var) and (type(var["number"]) is list) and (len(var["number"]) == 1) :
            var["number"] = var["number"][0]
        if ("number" in var) and (type(var["number"]) is str) :
            var["number"] = json.loads(var["number"])
        return var

    def _extract_variable_values(self, var, parent = None) :
        if "hasValues" in var:
            valuestr = var["hasValues"]
            values = json.loads(valuestr)
            if type(values) is dict and "base64_zlib" in values:
                values = unzip_string(values["base64_zlib"])
                var["hasValues"] = values
        return var
    
    def _changes_to_json(self, change, parent=None):
        newChange = {}
        if "name" in change:
            newChange[change["name"]] = change.get("notes", [])
            return newChange
        return None


from pylipd.globals.urls import ONTONS, NAMESPACES

QUERY_DSNAME = """
    SELECT ?dsname WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dsname
    }
"""

QUERY_DSID = """
    SELECT ?dsid WHERE {
        ?ds a le:Dataset .
        OPTIONAL{?ds le:hasDatasetId ?dsid}
    }
"""

QUERY_UNIQUE_ARCHIVE_TYPE = """
    SELECT distinct ?archiveType WHERE {
        ?ds a le:Dataset .
        ?ds le:hasArchiveType ?archiveTypeObj .
        ?archiveTypeObj rdfs:label ?archiveType .
    }
"""

QUERY_ENSEMBLE_TABLE_SHORT = """
    PREFIX le: <http://linked.earth/ontology#>

    SELECT ?datasetName ?ensembleTable ?ensembleVariableName ?ensembleVariableValues ?ensembleVariableUnits ?ensembleDepthName ?ensembleDepthValues ?ensembleDepthUnits ?notes 
    WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?datasetName .
            FILTER regex(?datasetName, "[dsname].*", "i").
    
        ?ds le:hasChronData ?chron .
        ?chron le:modeledBy ?model .
        ?model le:hasEnsembleTable ?ensembleTable .
            OPTIONAL{?ensembleTable le:hasNotes ?notes}
        
        ?ensembleTable le:hasVariable ?ensvar .
        ?ensvar le:hasName ?ensembleVariableName .
            FILTER (regex(lcase(?ensembleVariableName), "year.*", "i") || regex(?ensembleVariableName, "age.*", "i")) .
        ?ensvar le:hasValues ?ensembleVariableValues
            OPTIONAL{
                ?ensvar le:hasUnits ?ensembleVariableUnitsObj .
                ?ensembleVariableUnitsObj rdfs:label ?ensembleVariableUnits .
            }
        
        ?ensembleTable le:hasVariable ?ensdepthvar .
        ?ensdepthvar le:hasName ?ensembleDepthName .
            FILTER regex(lcase(?ensembleDepthName), "[ensembleDepthVarName].*").
        ?ensdepthvar le:hasValues ?ensembleDepthValues .
            OPTIONAL{
                ?ensdepthvar le:hasUnits ?ensembleDepthUnitsObj .
                ?ensembleDepthUnitsObj rdfs:label ?ensembleDepthUnits .
            }
    }
"""

QUERY_ENSEMBLE_TABLE = """
    PREFIX le: <http://linked.earth/ontology#>

    SELECT ?datasetName ?ensembleTable ?ensembleVariableName ?ensembleVariableValues ?ensembleVariableUnits ?ensembleDepthName ?ensembleDepthValues ?ensembleDepthUnits ?notes ?methodobj ?methods
    WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?datasetName .
            FILTER regex(?datasetName, "[dsname].*", "i").
    
        ?ds le:hasChronData ?chron .
        ?chron le:modeledBy ?model .
        ?model le:hasEnsembleTable ?ensembleTable .
            OPTIONAL{?ensembleTable le:hasNotes ?notes}
        
        ?ensembleTable le:hasVariable ?ensvar .
        ?ensvar le:hasName ?ensembleVariableName .
            FILTER regex(lcase(?ensembleVariableName), "[ensembleVarName].*", "i").
        ?ensvar le:hasValues ?ensembleVariableValues
            OPTIONAL{
                ?ensvar le:hasUnits ?ensembleVariableUnitsObj .
                ?ensembleVariableUnitsObj rdfs:label ?ensembleVariableUnits .
            }
        
        ?ensembleTable le:hasVariable ?ensdepthvar .
        ?ensdepthvar le:hasName ?ensembleDepthName .
            FILTER regex(lcase(?ensembleDepthName), "[ensembleDepthVarName].*", "i").
        ?ensdepthvar le:hasValues ?ensembleDepthValues .
            OPTIONAL{
                ?ensdepthvar le:hasUnits ?ensembleDepthUnitsObj .
                ?ensembleDepthUnitsObj rdfs:label ?ensembleDepthUnits .
            }
    }
"""


QUERY_BIBLIO = """
    SELECT ?dsname ?title (GROUP_CONCAT(?authorName;separator=" and ") as ?authors) 
    ?doi ?pubyear ?year ?journal ?volume ?issue ?pages ?type ?publisher ?report ?citeKey ?edition ?institution ?url ?url2
    WHERE { 
        ?ds a le:Dataset .
        ?ds le:hasName ?dsname .
        ?ds le:hasPublication ?pub .
        OPTIONAL{?pub le:hasDOI ?doi .}
        OPTIONAL{
            ?pub le:hasAuthor ?author .
            ?author le:hasName ?authorName .
        }
        OPTIONAL{?pub le:publicationYear ?pubyear .}
        OPTIONAL{?pub le:hasYear ?year .}
        OPTIONAL{?pub le:hasTitle ?title .}
        OPTIONAL{?pub le:hasJournal ?journal .}
        OPTIONAL{?pub le:hasVolume ?volume .}
        OPTIONAL{?pub le:hasIssue ?issue .}
        OPTIONAL{?pub le:hasPages ?pages .}
        OPTIONAL{?pub le:hasType ?type .}
        OPTIONAL{?pub le:hasPublisher ?publisher .}
        OPTIONAL{?pub le:hasReport ?report .}
        OPTIONAL{?pub le:hasCiteKey ?citeKey .}
        OPTIONAL{?pub le:hasEdition ?edition .}
        OPTIONAL{?pub le:hasInstitution ?institution .}
        OPTIONAL{?pub le:hasLink ?url .}
        OPTIONAL{?pub le:hasUrl ?url2 .}
    }
    GROUP BY ?pub ?dsname ?title ?doi ?year ?pubyear ?journal ?volume ?issue ?pages ?type ?publisher ?report ?citeKey ?edition ?institution ?url ?url2
"""

QUERY_DISTINCT_VARIABLE="""
    PREFIX le: <http://linked.earth/ontology#>
    
    SELECT distinct ?variableName 
    WHERE {
        ?uri le:hasName ?variableName .
        ?uri le:hasVariableId ?TSID
    }
    

"""

QUERY_DISTINCT_PROXY = """
    PREFIX le: <http://linked.earth/ontology#>
    
    SELECT distinct ?proxy 
    WHERE {
        OPTIONAL{?uri le:hasProxy ?proxyObj .
                 ?proxyObj rdfs:label ?proxy .}
        ?uri le:hasVariableId ?TSID
    }
    

"""

QUERY_VARIABLE = """
    PREFIX le: <http://linked.earth/ontology#>

    SELECT ?uri ?TSID ?variableName 
    WHERE {
        ?uri le:hasName ?variableName .
        ?uri le:hasVariableId ?TSID
    }
"""

QUERY_VARIABLE_GRAPH = """
    PREFIX le: <http://linked.earth/ontology#>

    CONSTRUCT {
        <[varid]> ?p1 ?o1 .
        <[varid]> ?pv1 ?v1 .
        ?s2 ?pv2 ?v2 .
    }
    WHERE {
        # level 1
        <[varid]> ?p1 ?o1 # get objects
            FILTER (
                (?p1 NOT IN (le:foundInTable, le:takenAtDepth)) &&
                isIRI(?o1)
            ) .
        <[varid]> ?pv1 ?v1  # get primitives
            FILTER (isLiteral(?v1)) .
        
        BIND (?o1 as ?s2) . # rename binding for readability

        # level 2
        ?s2 ?pv2 ?v2 
            FILTER (isLiteral(?v2)) .
    }
"""

QUERY_ALL_VARIABLES_GRAPH = """
    PREFIX le: <http://linked.earth/ontology#>

    INSERT {
        GRAPH ?var
        {
            ?var ?pv1 ?v1 .
            ?var ?p1 ?o1 .        
            ?o1 ?pv2 ?v2 .
            ?o1 ?p2 ?o2 .
            ?o2 ?pv3 ?v3 .
            ?var le:foundInTable ?table .
            ?var le:foundInDataset ?ds .
            ?var le:foundInDatasetName ?dsname
        }
    }
    WHERE {
        ?table le:hasVariable ?var .
        {
            {
                # level 1
                ?var le:foundInDataset ?ds .
                ?ds le:hasName ?dsname .
                ?var ?pv1 ?v1  # get primitives
                    FILTER (isLiteral(?v1)) .
            }
            UNION
            {
                # level 2
                {
                    ?var ?p1 ?o1
                        FILTER (?p1 NOT IN (le:foundInTable, le:foundInDataset)) .
                    ?o1 ?pv2 ?v2 
                        FILTER (isLiteral(?v2)) .
                } .
            }
            #UNION
            #{
            #    # level 3
            #    {
            #        ?var ?p1 ?o1
            #            FILTER (?p1 NOT IN (le:foundInTable, le:foundInDataset)) .
            #        ?o1 ?p2 ?o2
            #            FILTER (?p1 NOT IN (le:foundInTable, le:foundInDataset)) .
            #        ?o2 ?pv3 ?v3
            #            FILTER (isLiteral(?v2)) .
            #    } .
            #}
        }
    }
"""


QUERY_FILTER_GEO = """
    PREFIX wgs84: <http://www.w3.org/2003/01/geo/wgs84_pos#>

    SELECT ?dsname WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dsname .
        ?ds le:hasLocation ?loc .
        ?loc le:hasLatitude ?lat .
        ?loc le:hasLongitude ?lon .
        FILTER ( ?lat >= [latMin] && ?lat < [latMax] && ?lon >= [lonMin] && ?lon < [lonMax] ) .
    }
"""

QUERY_FILTER_ARCHIVE_TYPE = """
    SELECT ?dsname WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dsname .
        ?ds le:hasArchiveType ?archiveTypeObj .
        ?archiveTypeObj rdfs:label ?archiveType .
        FILTER regex(?archiveType, "[archiveType].*", "i")
    }
"""

QUERY_FILTER_DATASET_NAME = """
    SELECT ?dsname WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dsname .
        FILTER regex(?dsname, "[datasetName].*", "i")
    }
"""

QUERY_FILTER_TIME = """
    SELECT ?dsname ?minage ?maxage WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dsname .
        
        ?ds le:hasPaleoData ?data .
        ?data le:hasMeasurementTable ?table .
        ?table le:hasVariable ?var .
        ?table le:hasVariable ?timevar .
        ?timevar le:hasName ?time_variableName .
        FILTER (regex(?time_variableName, "year.*") || regex(?time_variableName, "age.*")) .
        ?timevar le:hasMinValue ?minage .
        ?timevar le:hasMaxValue ?maxage .
}
"""

QUERY_FILTER_VARIABLE_NAME = """
    SELECT ?uri ?dsuri ?dsname ?tableuri ?id ?name WHERE {
        ?uri le:hasVariableId ?id .
        ?uri le:hasName ?name .
        FILTER regex(?name, "[name].*", "i") .
        ?uri le:foundInDataset ?dsuri .
        ?uri le:foundInDatasetName ?dataSetName .
        ?uri le:foundInTable ?tableuri .
    }
"""

QUERY_FILTER_VARIABLE_PROXY = """
    SELECT ?uri ?dsuri ?dsname ?tableuri ?id ?proxy WHERE {
        ?uri le:hasVariableId ?id .
        ?uri le:hasProxy ?proxyObj .
        ?proxyObj rdfs:label ?proxy .
        FILTER regex(?proxy, "[proxy].*", "i") .
        ?uri le:foundInDataset ?dsuri .
        ?uri le:foundInDatasetName ?dataSetName .
        ?uri le:foundInTable ?tableuri .
    }
"""

QUERY_FILTER_VARIABLE_RESOLUTION = """
    SELECT ?uri ?dsuri ?dsname ?tableuri ?id ?v WHERE {
        ?uri le:hasVariableId ?id .
        ?uri le:hasResolution ?res .
        ?res le:has[stat]Value ?v .
        FILTER(?v<[value]) .
        ?uri le:foundInDataset ?dsuri .
        ?uri le:foundInDatasetName ?dataSetName .
        ?uri le:foundInTable ?tableuri .        
    }
"""


QUERY_TIMESERIES_ESSENTIALS_PALEO ="""
    PREFIX wgs84: <http://www.w3.org/2003/01/geo/wgs84_pos#>
    PREFIX le: <http://linked.earth/ontology#>

    SELECT ?dataSetName ?archiveType ?geo_meanLat ?geo_meanLon ?geo_meanElev 
    ?paleoData_variableName ?paleoData_values ?paleoData_units 
    ?paleoData_proxy ?paleoData_proxyGeneral ?time_variableName ?time_values 
    ?time_units ?depth_variableName ?depth_values ?depth_units WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dataSetName .
            FILTER regex(?dataSetName, "[dsname].*", "i").
        
        OPTIONAL{
            ?ds le:hasArchiveType ?archiveTypeObj .
            ?archiveTypeObj rdfs:label ?archiveType .
        }
        
        ?ds le:hasLocation ?loc .
        OPTIONAL { ?loc le:hasLatitude ?geo_meanLat } .
        OPTIONAL { ?loc le:hasLongitude ?geo_meanLon } .
        OPTIONAL { ?loc le:hasElevation ?geo_meanElev } .
        
        ?ds le:hasPaleoData ?data .
        ?data le:hasMeasurementTable ?table .
        ?table le:hasVariable ?var .
        
        ?var le:hasName ?paleoData_variableName .
        FILTER (!regex(?paleoData_variableName, "year.*") && !regex(?paleoData_variableName, "age.*") && !regex(?paleoData_variableName, "depth.*")) .
   		
        ?var le:hasValues ?paleoData_values .
        OPTIONAL{
            ?var le:hasUnits ?paleoData_unitsObj .
            ?paleoData_unitsObj rdfs:label ?paleoData_units .
        }
        OPTIONAL{
            ?var le:hasProxy ?paleoData_proxyObj .
            ?paleoData_proxyObj rdfs:label ?paleoData_proxy .
        }
        OPTIONAL{
            ?var le:hasProxyGeneral ?paleoData_proxyGeneralObj .
            ?paleoData_proxyGeneralObj rdfs:label ?paleoData_proxyGeneral .
        }
        
        
        OPTIONAL{
            ?table le:hasVariable ?timevar .
            ?timevar le:hasName ?time_variableName .
                FILTER (regex(?time_variableName, "year.*") || regex(?time_variableName, "age.*")) .
            ?timevar le:hasValues ?time_values .
            OPTIONAL{
                ?timevar le:hasUnits ?time_unitsObj .
                ?time_unitsObj rdfs:label ?time_units .
            }
        }
        
        OPTIONAL{
            ?table le:hasVariable ?depthvar .
            ?depthvar le:hasName ?depth_variableName .
                FILTER (regex(?depth_variableName, "depth.*")) .
            ?depthvar le:hasValues ?depth_values .
            OPTIONAL{
                ?depthvar le:hasUnits ?depth_unitsObj .
                ?depth_unitsObj rdfs:label ?depth_units .
            }
        }
        
    }
"""

QUERY_TIMESERIES_ESSENTIALS_CHRON ="""
    PREFIX wgs84: <http://www.w3.org/2003/01/geo/wgs84_pos#>
    PREFIX le: <http://linked.earth/ontology#>

    SELECT ?dataSetName ?archiveType ?geo_meanLat ?geo_meanLon ?geo_meanElev 
    ?chronData_variableName ?chronData_values ?chronData_units 
    ?time_variableName ?time_values 
    ?time_units ?depth_variableName ?depth_values ?depth_units WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dataSetName .
            FILTER regex(?dataSetName, "[dsname].*", "i").
        
        OPTIONAL{
            ?ds le:hasArchiveType ?archiveTypeObj .
            ?archiveTypeObj rdfs:label ?archiveType .
        }
        
        ?ds le:hasLocation ?loc .
        OPTIONAL { ?loc le:hasLatitude ?geo_meanLat } .
        OPTIONAL { ?loc le:hasLongitude ?geo_meanLon } .
        OPTIONAL { ?loc le:hasElevation ?geo_meanElev } .
        
        ?ds le:hasChronData ?data .
        ?data le:hasMeasurementTable ?table .
        ?table le:hasVariable ?var .
        ?var le:hasName ?chronData_variableName .
   		
        ?var le:hasValues ?chronData_values .
        OPTIONAL{
            ?var le:hasUnits ?chronData_unitsObj .
            ?chronData_unitsObj rdfs:label ?chronData_units .
        }
        
        OPTIONAL{?table le:hasVariable ?timevar .
        ?timevar le:hasName ?time_variableName .
            FILTER (regex(?time_variableName, "year.*") || regex(?time_variableName, "age.*")) .
        ?timevar le:hasValues ?time_values .
            OPTIONAL{
                ?timevar le:hasUnits ?time_unitsObj .
                ?time_unitsObj rdfs:label ?time_units .
            }
        }
        
        OPTIONAL{?table le:hasVariable ?depthvar .
        ?depthvar le:hasName ?depth_variableName .
            FILTER (regex(?depth_variableName, "depth.*")) .
        ?depthvar le:hasValues ?depth_values .
            OPTIONAL{
                ?depthvar le:hasUnits ?depth_unitsObj .
                ?depth_unitsObj rdfs:label ?depth_units .
            }
        }
    }
"""

QUERY_VARIABLE_PROPERTIES="""
    PREFIX le: <http://linked.earth/ontology#>
    SELECT  DISTINCT ?property where {
    
    ?ds a le:Dataset .
    
    {?ds le:hasPaleoData ?data .
    ?data le:hasMeasurementTable ?table .
    ?table le:hasVariable ?var .
    ?var ?property ?value .}
    
    UNION
    
    {OPTIONAL{?ds le:hasChronData ?data1 .
    ?data1 le:hasMeasurementTable ?table1 .
    ?table1 le:hasVariable ?var1 .
    ?var1 ?property ?value1 .}}
    
    }
"""

## At the LiPDSeries level

QUERY_LiPDSERIES_PROPERTIES="""
    SELECT DISTINCT ?p WHERE {
        ?uri ?p ?v .}
    """

QUERY_DATASET_PROPERTIES="""
    PREFIX le: <http://linked.earth/ontology#>
    SELECT DISTINCT ?property where {
    ?ds a le:Dataset .
    ?ds ?property ?value .
    }
"""

QUERY_MODEL_PROPERTIES="""
    PREFIX le: <http://linked.earth/ontology#>
    SELECT  DISTINCT ?property where {
    
    ?ds a le:Dataset .
    
    {OPTIONAL{?ds le:hasPaleoData ?data .
              ?data le:modeledBy ?paleomodel .
              ?paleomodel ?property ?value .}}
    
    UNION
    
    {OPTIONAL{?ds le:hasChronData ?chron .
              ?chron le:modeledBy ?chronmodel .
              ?chronmodel ?property ?value .}}
    
    }
"""

QUERY_VARIABLE_ESSENTIALS="""
    PREFIX le: <http://linked.earth/ontology#>
    SELECT ?dataSetName ?archiveType ?name ?TSID ?values ?units ?proxy where {
        ?var le:hasName ?name .
        ?var le:foundInDatasetName ?dataSetName .
            #FILTER regex(?dataSetName, "[dsname].*", "i").
            
        OPTIONAL{?var le:hasVariableId ?TSID .}
        ?var le:hasValues ?values .
        OPTIONAL{
            ?var le:hasUnits ?unitsObj .
            ?unitsObj rdfs:label ?units .
        }
        OPTIONAL{
            ?var le:hasArchiveType ?archiveTypeObj .
            ?archiveTypeObj rdfs:label ?archiveType .
        }
        OPTIONAL{
            ?var le:hasProxy ?paleoData_proxyObj .
            ?paleoData_proxyObj rdfs:label ?paleoData_proxy .
        }    
    }
"""

QUERY_LOCATION ="""
    PREFIX wgs84: <http://www.w3.org/2003/01/geo/wgs84_pos#>
    PREFIX le: <http://linked.earth/ontology#>

    SELECT ?dataSetName ?geo_meanLat ?geo_meanLon ?geo_meanElev WHERE {
        ?ds a le:Dataset .
        ?ds le:hasName ?dataSetName .
            FILTER regex(?dataSetName, "[dsname].*", "i").
        
        ?ds le:hasLocation ?loc .
        OPTIONAL { ?loc le:hasLatitude ?geo_meanLat } .
        OPTIONAL { ?loc le:hasLongitude ?geo_meanLon } .
        OPTIONAL { ?loc le:hasElevation ?geo_meanElev } .
    }
"""

QUERY_FILTER_COMPILATION="""
        SELECT DISTINCT ?dataSetName WHERE {
            ?ds a le:Dataset .
            ?ds le:hasName ?dataSetName .
        
            ?ds le:hasPaleoData ?data .
            ?data le:hasMeasurementTable ?table .
            ?table le:hasVariable ?var .
            
            ?var le:partOfCompilation ?compilation . 
            ?compilation le:hasName ?compilationName .
            FILTER regex(?compilationName, "[compilationName].*", "i")}
            
    """

QUERY_COMPILATION_NAME="""
        PREFIX le: <http://linked.earth/ontology#>
        
        SELECT DISTINCT ?compilationName WHERE {
            ?var a le:Variable .
            ?var le:partOfCompilation ?compilation . 
            ?compilation le:hasName ?compilationName .}
"""




##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid

class Uncertainty:

    def __init__(self):
        self.uncertaintyType: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#Uncertainty"
        self.id = self.ns + "/" + uniqid("Uncertainty.")

    @staticmethod
    def from_data(id, data) -> 'Uncertainty':
        self = Uncertainty()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasUncertaintyType":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.uncertaintyType = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if self.uncertaintyType:
            value_obj = self.uncertaintyType
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasUncertaintyType"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match("\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        data = {
            "@id": self.id
        }

        if self.uncertaintyType:
            value_obj = self.uncertaintyType
            obj = value_obj
            data["uncertaintyType"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'Uncertainty':
        self = Uncertainty()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "uncertaintyType":
                    value = pvalue
                    obj = value
                    self.uncertaintyType = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        return self.misc

    def add_non_standard_property(self, key, value):
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getUncertaintyType(self) -> str:
        return self.uncertaintyType

    def setUncertaintyType(self, uncertaintyType:str):
        self.uncertaintyType = uncertaintyType
    



##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid

class PhysicalSample:
    """Auto-generated LinkedEarth class representing `PhysicalSample`."""
    def __init__(self):
        """Initialize a new PhysicalSample instance."""
        
        self.housedAt: str = None
        self.iGSN: str = None
        self.name: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#PhysicalSample"
        self.id = self.ns + "/" + uniqid("PhysicalSample.")

    @staticmethod
    def from_data(id, data) -> 'PhysicalSample':
        """Instantiate `PhysicalSample` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        PhysicalSample
            The populated `PhysicalSample` instance.
        """
        self = PhysicalSample()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasIGSN":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.iGSN = obj
        
            elif key == "housedAt":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.housedAt = obj
        
            elif key == "name":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.name = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if self.housedAt:
            value_obj = self.housedAt
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["housedAt"] = [obj]
                

        if self.iGSN:
            value_obj = self.iGSN
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasIGSN"] = [obj]
                

        if self.name:
            value_obj = self.name
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["name"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if self.housedAt:
            value_obj = self.housedAt
            obj = value_obj
            data["housedat"] = obj

        if self.iGSN:
            value_obj = self.iGSN
            obj = value_obj
            data["hasidentifier"] = obj

        if self.name:
            value_obj = self.name
            obj = value_obj
            data["hasname"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'PhysicalSample':
        """Instantiate `PhysicalSample` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        PhysicalSample
            The populated `PhysicalSample` instance.
        """
        self = PhysicalSample()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "hasidentifier":
                    value = pvalue
                    obj = value
                    self.iGSN = obj
            elif key == "hasname":
                    value = pvalue
                    obj = value
                    self.name = obj
            elif key == "housedat":
                    value = pvalue
                    obj = value
                    self.housedAt = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getHousedAt(self) -> str:
        """Get housedAt.

        Returns
        -------
        str
            The current value of housedAt.
        """
        return self.housedAt

    def setHousedAt(self, housedAt:str):
        """Set housedAt.

        Parameters
        ----------
        housedAt : str
            The value to assign.
        """
        assert isinstance(housedAt, str), f"Error: '{housedAt}' is not of type str"
        self.housedAt = housedAt
    
    def getIGSN(self) -> str:
        """Get iGSN.

        Returns
        -------
        str
            The current value of iGSN.
        """
        return self.iGSN

    def setIGSN(self, iGSN:str):
        """Set iGSN.

        Parameters
        ----------
        iGSN : str
            The value to assign.
        """
        assert isinstance(iGSN, str), f"Error: '{iGSN}' is not of type str"
        self.iGSN = iGSN
    
    def getName(self) -> str:
        """Get name.

        Returns
        -------
        str
            The current value of name.
        """
        return self.name

    def setName(self, name:str):
        """Set name.

        Parameters
        ----------
        name : str
            The value to assign.
        """
        assert isinstance(name, str), f"Error: '{name}' is not of type str"
        self.name = name
    



##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid

class Location:
    """Auto-generated LinkedEarth class representing `Location`."""
    def __init__(self):
        """Initialize a new Location instance."""
        
        self.continent: str = None
        self.coordinates: str = None
        self.coordinatesFor: None = None
        self.country: str = None
        self.countryOcean: str = None
        self.description: str = None
        self.elevation: str = None
        self.geometryType: str = None
        self.latitude: str = None
        self.locationName: str = None
        self.locationType: str = None
        self.longitude: str = None
        self.notes: str = None
        self.ocean: str = None
        self.siteName: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#Location"
        self.id = self.ns + "/" + uniqid("Location.")

    @staticmethod
    def from_data(id, data) -> 'Location':
        """Instantiate `Location` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        Location
            The populated `Location` instance.
        """
        self = Location()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "coordinates":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.coordinates = obj
        
            elif key == "coordinatesFor":
                for val in value:
                    obj = val["@id"]                        
                    self.coordinatesFor = obj
        
            elif key == "hasContinent":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.continent = obj
        
            elif key == "hasCountry":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.country = obj
        
            elif key == "hasCountryOcean":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.countryOcean = obj
        
            elif key == "hasDescription":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.description = obj
        
            elif key == "hasElevation":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.elevation = obj
        
            elif key == "hasGeometryType":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.geometryType = obj
        
            elif key == "hasLatitude":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.latitude = obj
        
            elif key == "hasLocationName":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.locationName = obj
        
            elif key == "hasLongitude":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.longitude = obj
        
            elif key == "hasNotes":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.notes = obj
        
            elif key == "hasOcean":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.ocean = obj
        
            elif key == "hasSiteName":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.siteName = obj
        
            elif key == "hasType":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.locationType = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if self.continent:
            value_obj = self.continent
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasContinent"] = [obj]
                

        if self.coordinates:
            value_obj = self.coordinates
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["coordinates"] = [obj]
                

        if self.coordinatesFor:
            value_obj = self.coordinatesFor
            obj = {
                "@id": value_obj,
                "@type": "uri"
            }
            data[self.id]["coordinatesFor"] = [obj]
                

        if self.country:
            value_obj = self.country
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasCountry"] = [obj]
                

        if self.countryOcean:
            value_obj = self.countryOcean
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasCountryOcean"] = [obj]
                

        if self.description:
            value_obj = self.description
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasDescription"] = [obj]
                

        if self.elevation:
            value_obj = self.elevation
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasElevation"] = [obj]
                

        if self.geometryType:
            value_obj = self.geometryType
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasGeometryType"] = [obj]
                

        if self.latitude:
            value_obj = self.latitude
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasLatitude"] = [obj]
                

        if self.locationName:
            value_obj = self.locationName
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasLocationName"] = [obj]
                

        if self.locationType:
            value_obj = self.locationType
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasType"] = [obj]
                

        if self.longitude:
            value_obj = self.longitude
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasLongitude"] = [obj]
                

        if self.notes:
            value_obj = self.notes
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasNotes"] = [obj]
                

        if self.ocean:
            value_obj = self.ocean
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasOcean"] = [obj]
                

        if self.siteName:
            value_obj = self.siteName
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasSiteName"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if self.continent:
            value_obj = self.continent
            obj = value_obj
            data["continent"] = obj

        if self.coordinates:
            value_obj = self.coordinates
            obj = value_obj
            data["coordinates"] = obj

        if self.coordinatesFor:
            value_obj = self.coordinatesFor
            obj = value_obj
            data["coordinatesFor"] = obj

        if self.country:
            value_obj = self.country
            obj = value_obj
            data["country"] = obj

        if self.countryOcean:
            value_obj = self.countryOcean
            obj = value_obj
            data["countryOcean"] = obj

        if self.description:
            value_obj = self.description
            obj = value_obj
            data["description"] = obj

        if self.elevation:
            value_obj = self.elevation
            obj = value_obj
            data["elevation"] = obj

        if self.geometryType:
            value_obj = self.geometryType
            obj = value_obj
            data["geometryType"] = obj

        if self.latitude:
            value_obj = self.latitude
            obj = value_obj
            data["latitude"] = obj

        if self.locationName:
            value_obj = self.locationName
            obj = value_obj
            data["locationName"] = obj

        if self.locationType:
            value_obj = self.locationType
            obj = value_obj
            data["type"] = obj

        if self.longitude:
            value_obj = self.longitude
            obj = value_obj
            data["longitude"] = obj

        if self.notes:
            value_obj = self.notes
            obj = value_obj
            data["notes"] = obj

        if self.ocean:
            value_obj = self.ocean
            obj = value_obj
            data["ocean"] = obj

        if self.siteName:
            value_obj = self.siteName
            obj = value_obj
            data["siteName"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'Location':
        """Instantiate `Location` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        Location
            The populated `Location` instance.
        """
        self = Location()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "continent":
                    value = pvalue
                    obj = value
                    self.continent = obj
            elif key == "coordinates":
                    value = pvalue
                    obj = value
                    self.coordinates = obj
            elif key == "coordinatesFor":
                    value = pvalue
                    obj = value
                    self.coordinatesFor = obj
            elif key == "country":
                    value = pvalue
                    obj = value
                    self.country = obj
            elif key == "countryOcean":
                    value = pvalue
                    obj = value
                    self.countryOcean = obj
            elif key == "description":
                    value = pvalue
                    obj = value
                    self.description = obj
            elif key == "elevation":
                    value = pvalue
                    obj = value
                    self.elevation = obj
            elif key == "geometryType":
                    value = pvalue
                    obj = value
                    self.geometryType = obj
            elif key == "latitude":
                    value = pvalue
                    obj = value
                    self.latitude = obj
            elif key == "locationName":
                    value = pvalue
                    obj = value
                    self.locationName = obj
            elif key == "longitude":
                    value = pvalue
                    obj = value
                    self.longitude = obj
            elif key == "notes":
                    value = pvalue
                    obj = value
                    self.notes = obj
            elif key == "ocean":
                    value = pvalue
                    obj = value
                    self.ocean = obj
            elif key == "siteName":
                    value = pvalue
                    obj = value
                    self.siteName = obj
            elif key == "type":
                    value = pvalue
                    obj = value
                    self.locationType = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getContinent(self) -> str:
        """Get continent.

        Returns
        -------
        str
            The current value of continent.
        """
        return self.continent

    def setContinent(self, continent:str):
        """Set continent.

        Parameters
        ----------
        continent : str
            The value to assign.
        """
        assert isinstance(continent, str), f"Error: '{continent}' is not of type str"
        self.continent = continent
    
    def getCoordinates(self) -> str:
        """Get coordinates.

        Returns
        -------
        str
            The current value of coordinates.
        """
        return self.coordinates

    def setCoordinates(self, coordinates:str):
        """Set coordinates.

        Parameters
        ----------
        coordinates : str
            The value to assign.
        """
        assert isinstance(coordinates, str), f"Error: '{coordinates}' is not of type str"
        self.coordinates = coordinates
    
    def getCoordinatesFor(self) -> object:
        """Get coordinatesFor.

        Returns
        -------
        object
            The current value of coordinatesFor.
        """
        return self.coordinatesFor

    def setCoordinatesFor(self, coordinatesFor:object):
        """Set coordinatesFor.

        Parameters
        ----------
        coordinatesFor : object
            The value to assign.
        """
        assert isinstance(coordinatesFor, object), f"Error: '{coordinatesFor}' is not of type object"
        self.coordinatesFor = coordinatesFor
    
    def getCountry(self) -> str:
        """Get country.

        Returns
        -------
        str
            The current value of country.
        """
        return self.country

    def setCountry(self, country:str):
        """Set country.

        Parameters
        ----------
        country : str
            The value to assign.
        """
        assert isinstance(country, str), f"Error: '{country}' is not of type str"
        self.country = country
    
    def getCountryOcean(self) -> str:
        """Get countryOcean.

        Returns
        -------
        str
            The current value of countryOcean.
        """
        return self.countryOcean

    def setCountryOcean(self, countryOcean:str):
        """Set countryOcean.

        Parameters
        ----------
        countryOcean : str
            The value to assign.
        """
        assert isinstance(countryOcean, str), f"Error: '{countryOcean}' is not of type str"
        self.countryOcean = countryOcean
    
    def getDescription(self) -> str:
        """Get description.

        Returns
        -------
        str
            The current value of description.
        """
        return self.description

    def setDescription(self, description:str):
        """Set description.

        Parameters
        ----------
        description : str
            The value to assign.
        """
        assert isinstance(description, str), f"Error: '{description}' is not of type str"
        self.description = description
    
    def getElevation(self) -> str:
        """Get elevation.

        Returns
        -------
        str
            The current value of elevation.
        """
        return self.elevation

    def setElevation(self, elevation:str):
        """Set elevation.

        Parameters
        ----------
        elevation : str
            The value to assign.
        """
        assert isinstance(elevation, str), f"Error: '{elevation}' is not of type str"
        self.elevation = elevation
    
    def getGeometryType(self) -> str:
        """Get geometryType.

        Returns
        -------
        str
            The current value of geometryType.
        """
        return self.geometryType

    def setGeometryType(self, geometryType:str):
        """Set geometryType.

        Parameters
        ----------
        geometryType : str
            The value to assign.
        """
        assert isinstance(geometryType, str), f"Error: '{geometryType}' is not of type str"
        self.geometryType = geometryType
    
    def getLatitude(self) -> str:
        """Get latitude.

        Returns
        -------
        str
            The current value of latitude.
        """
        return self.latitude

    def setLatitude(self, latitude:str):
        """Set latitude.

        Parameters
        ----------
        latitude : str
            The value to assign.
        """
        assert isinstance(latitude, str), f"Error: '{latitude}' is not of type str"
        self.latitude = latitude
    
    def getLocationName(self) -> str:
        """Get locationName.

        Returns
        -------
        str
            The current value of locationName.
        """
        return self.locationName

    def setLocationName(self, locationName:str):
        """Set locationName.

        Parameters
        ----------
        locationName : str
            The value to assign.
        """
        assert isinstance(locationName, str), f"Error: '{locationName}' is not of type str"
        self.locationName = locationName
    
    def getLocationType(self) -> str:
        """Get locationType.

        Returns
        -------
        str
            The current value of locationType.
        """
        return self.locationType

    def setLocationType(self, locationType:str):
        """Set locationType.

        Parameters
        ----------
        locationType : str
            The value to assign.
        """
        assert isinstance(locationType, str), f"Error: '{locationType}' is not of type str"
        self.locationType = locationType
    
    def getLongitude(self) -> str:
        """Get longitude.

        Returns
        -------
        str
            The current value of longitude.
        """
        return self.longitude

    def setLongitude(self, longitude:str):
        """Set longitude.

        Parameters
        ----------
        longitude : str
            The value to assign.
        """
        assert isinstance(longitude, str), f"Error: '{longitude}' is not of type str"
        self.longitude = longitude
    
    def getNotes(self) -> str:
        """Get notes.

        Returns
        -------
        str
            The current value of notes.
        """
        return self.notes

    def setNotes(self, notes:str):
        """Set notes.

        Parameters
        ----------
        notes : str
            The value to assign.
        """
        assert isinstance(notes, str), f"Error: '{notes}' is not of type str"
        self.notes = notes
    
    def getOcean(self) -> str:
        """Get ocean.

        Returns
        -------
        str
            The current value of ocean.
        """
        return self.ocean

    def setOcean(self, ocean:str):
        """Set ocean.

        Parameters
        ----------
        ocean : str
            The value to assign.
        """
        assert isinstance(ocean, str), f"Error: '{ocean}' is not of type str"
        self.ocean = ocean
    
    def getSiteName(self) -> str:
        """Get siteName.

        Returns
        -------
        str
            The current value of siteName.
        """
        return self.siteName

    def setSiteName(self, siteName:str):
        """Set siteName.

        Parameters
        ----------
        siteName : str
            The value to assign.
        """
        assert isinstance(siteName, str), f"Error: '{siteName}' is not of type str"
        self.siteName = siteName
    


import pandas as pd
import json
from pylipd.classes.variable import Variable

### START TEMPLATE FUNCTION ###
# Special Functions manually added for DataTable class
def getDataFrame(self, use_standard_names=False) -> pd.DataFrame:
    cols = []
    for v in self.variables:
        colname = v.getName()
        if use_standard_names and v.getStandardVariable() is not None:
            colname = v.getStandardVariable().getLabel()
        cols.append(colname)
    
    df = pd.DataFrame(columns=cols)
    for v in self.variables:
        colname = v.getName()
        if use_standard_names and v.getStandardVariable() is not None:
            colname = v.getStandardVariable().getLabel()
        df[colname] = json.loads(v.getValues())
    
    # Create metadata as a dictionary and add to dataframe attr
    df.attrs = {}
    for v in self.variables:
        colname = v.getName()
        if use_standard_names and v.getStandardVariable() is not None:
            colname = v.getStandardVariable().getLabel()
        df.attrs[colname] = v.to_json()
        del df.attrs[colname]["hasValues"]

    return df
### END TEMPLATE FUNCTION ###


### START TEMPLATE FUNCTION ###
def setDataFrame(self, df: pd.DataFrame):
    # Create new set of variable objects using the metadata
    self.variables = []
    for colname in df.attrs:
        v = Variable.from_json(df.attrs[colname])
        v.setValues(json.dumps(df[colname].to_list()))
        self.addVariable(v)
### END TEMPLATE FUNCTION ###


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 17 22:34:58 2023

@author: deborahkhider

Tests for pylipd.LiPDSeries

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}
Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
"""
import pytest
from pylipd.lipd_series import LiPDSeries

class TestLoad():
    
    def test_load_t0(self,odp846):
        D =odp846
        lipd_series = D.to_lipd_series()
        
class TestGetAll():
    
    def test_variables_t0(self,odp846):
        D =odp846
        S = D.to_lipd_series()
        df = S.get_all_variables()
    
    def test_variable_names_t0(self,pages2k):
        D=pages2k
        S = D.to_lipd_series()
        names = S.get_all_variable_names()
    
    def test_timeseries_essentials_t0(self, pages2k):
        D=pages2k
        S = D.to_lipd_series()
        names = S.get_timeseries_essentials()
    
    def test_proxy_t0(self, pages2k):
        D=pages2k
        S = D.to_lipd_series()
        names = S.get_all_proxy()
    
    def test_variable_t0(self,pages2k):
        D=pages2k
        S = D.to_lipd_series()
        l = S.get_variable_properties()
        
class TestFiler():
    
    def test_name_t0(self,pages2k):
        D=pages2k
        S = D.to_lipd_series()
        Sfiltered = S.filter_by_name('temperature')
        df=Sfiltered.get_timeseries_essentials()
        assert len(df.index)==11
    
    def test_proxy_t0(self,pages2k):
        D=pages2k
        S = D.to_lipd_series()
        Sfiltered = S.filter_by_proxy('ring width')
        v = Sfiltered.get_all_proxy()
        assert len(v)==1
    
    @pytest.mark.parametrize('stats',['Mean','Median','Min','Max'])
    def test_resolution_t0(self,stats,pages2k):
        D=pages2k
        S = D.to_lipd_series()
        Sfiltered = S.filter_by_resolution(threshold = 10,stats=stats)
        

        


import copy
import re
import uuid

from pandas import DataFrame
from rdflib.plugins.sparql.processor import SPARQLResult

import zlib, json, base64

def ucfirst(s):
    return s[0].upper() + s[1:]

def lcfirst(s):
    return s[0].lower() + s[1:]

def camelCase(id) :
    term = ""
    for subid in re.split(r"\s+", id): 
        term += ucfirst(subid)
    return term

def unCamelCase(id) :
    regex = r"(?<=[a-z])(?=[A-Z]) | (?<=[A-Z])(?=[A-Z][a-z])"
    a = re.split(regex, id)
    return " ".join(a).lower()

def fromCamelCase(str) :
    return ucfirst(str)
    #return ucfirst(str.replace(r"([^A-Z])([A-Z])"", "$1_$2", str))

def escape( str ):
    # str = str.replace("&", "&amp;")
    # str = str.replace("<", "&lt;")
    # str = str.replace(">", "&gt;")
    # str = str.replace("\"", "\\\"")
    # str = str.replace("'", "\\'")
    # str = str.replace("\n", " ")
    # str = str.replace("\r", " ")
    # str = re.sub(r"\\$", "", str)
    return str

def sanitizeId(id):
    return re.sub(r"[^a-zA-Z0-9\-_\.]", "_", id)

def uniqid(prefix='', more_entropy=False):
    """
    Generate a truly unique identifier using UUID4.
    
    Parameters:
    -----------
    prefix : str
        Optional prefix to prepend to the UUID
    more_entropy : bool
        If True, generates a longer UUID with additional entropy
        
    Returns:
    --------
    str
        A unique identifier string
    """
    if more_entropy:
        # Generate two UUIDs for extra entropy and combine them
        uuid1 = str(uuid.uuid4()).replace('-', '')
        uuid2 = str(uuid.uuid4()).replace('-', '')[:8]  # Take first 8 chars of second UUID
        the_uniqid = uuid1 + uuid2
    else:
        # Standard UUID4 without hyphens
        the_uniqid = str(uuid.uuid4()).replace('-', '')
    
    # Add prefix if provided
    return (prefix if prefix else '') + the_uniqid

def zip_string(string):
    return base64.b64encode(
        zlib.compress(string.encode('utf-8'))
    ).decode('ascii')

def unzip_string(string):
    try:
        return  zlib.decompress(base64.b64decode(string))
    except:
        raise RuntimeError("Could not decode/unzip the contents")

def expand_schema(schema) :
    xschema = {}
    for key,props in schema.items() :
        # Add core schema too
        xschema[key] = copy.copy(props)
        for lipdkey,pdetails in props.items() :
            if not type(pdetails) is dict:
                continue
            
            if (("alternates" in pdetails)) :
                for altkey in pdetails["alternates"]: 
                    xschema[key][altkey] = pdetails
    xschema["__expanded"] = True
    return xschema


def sparql_results_to_df(results: SPARQLResult) -> DataFrame:
    """
    Export results from an rdflib SPARQL query into a `pandas.DataFrame`,
    using Python types. See https://github.com/RDFLib/rdflib/issues/1179.
    """
    return DataFrame(
        data=([None if x is None else x.toPython() for x in row] for row in results),
        columns=[str(x) for x in results.vars],
    )



ONTONS = "http://linked.earth/ontology#"
NSURL = "http://linked.earth/lipd"
DATAURL = "https://data.mint.isi.edu/files/lipd"

ARCHIVEURL = "http://linked.earth/ontology/archive"
PROXYURL = "http://linked.earth/ontology/proxy"
UNITSURL = "http://linked.earth/ontology/units"
VARIABLEURL = "http://linked.earth/ontology/variables"

DEFAULT_GRAPH_URI = "http://www.openrdf.org/schema/sesame#nil"

NAMESPACES = {
    "wgs84": "http://www.w3.org/2003/01/geo/wgs84_pos#",
    "le_archive": ARCHIVEURL,
    "le_proxy": PROXYURL,
    "le_units": UNITSURL,
    "le_variables": VARIABLEURL
}





##############################
# Auto-generated. Do not Edit
##############################

import re
from pylipd.utils import uniqid
from pylipd.classes.interpretationseasonality import InterpretationSeasonality
from pylipd.classes.interpretationvariable import InterpretationVariable

class Interpretation:
    """Auto-generated LinkedEarth class representing `Interpretation`."""
    def __init__(self):
        """Initialize a new Interpretation instance."""
        
        self.basis: str = None
        self.direction: str = None
        self.local: str = None
        self.mathematicalRelation: str = None
        self.notes: str = None
        self.rank: str = None
        self.scope: str = None
        self.seasonality: InterpretationSeasonality = None
        self.seasonalityGeneral: InterpretationSeasonality = None
        self.seasonalityOriginal: InterpretationSeasonality = None
        self.variable: InterpretationVariable = None
        self.variableDetail: str = None
        self.variableGeneral: str = None
        self.variableGeneralDirection: str = None
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#Interpretation"
        self.id = self.ns + "/" + uniqid("Interpretation.")

    @staticmethod
    def from_data(id, data) -> 'Interpretation':
        """Instantiate `Interpretation` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        Interpretation
            The populated `Interpretation` instance.
        """
        self = Interpretation()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasBasis":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.basis = obj
        
            elif key == "hasDirection":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.direction = obj
        
            elif key == "hasMathematicalRelation":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.mathematicalRelation = obj
        
            elif key == "hasNotes":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.notes = obj
        
            elif key == "hasRank":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.rank = obj
        
            elif key == "hasScope":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.scope = obj
        
            elif key == "hasSeasonality":
                for val in value:
                    obj = InterpretationSeasonality.from_synonym(re.sub("^.*?#", "", val["@id"]))
                                    
                    self.seasonality = obj
        
            elif key == "hasSeasonalityGeneral":
                for val in value:
                    obj = InterpretationSeasonality.from_synonym(re.sub("^.*?#", "", val["@id"]))
                                    
                    self.seasonalityGeneral = obj
        
            elif key == "hasSeasonalityOriginal":
                for val in value:
                    obj = InterpretationSeasonality.from_synonym(re.sub("^.*?#", "", val["@id"]))
                                    
                    self.seasonalityOriginal = obj
        
            elif key == "hasVariable":
                for val in value:
                    obj = InterpretationVariable.from_synonym(re.sub("^.*?#", "", val["@id"]))
                                    
                    self.variable = obj
        
            elif key == "hasVariableDetail":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.variableDetail = obj
        
            elif key == "hasVariableGeneral":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.variableGeneral = obj
        
            elif key == "hasVariableGeneralDirection":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.variableGeneralDirection = obj
        
            elif key == "isLocal":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.local = obj
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if self.basis:
            value_obj = self.basis
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasBasis"] = [obj]
                

        if self.direction:
            value_obj = self.direction
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasDirection"] = [obj]
                

        if self.local:
            value_obj = self.local
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["isLocal"] = [obj]
                

        if self.mathematicalRelation:
            value_obj = self.mathematicalRelation
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasMathematicalRelation"] = [obj]
                

        if self.notes:
            value_obj = self.notes
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasNotes"] = [obj]
                

        if self.rank:
            value_obj = self.rank
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasRank"] = [obj]
                

        if self.scope:
            value_obj = self.scope
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasScope"] = [obj]
                

        if self.seasonality:
            value_obj = self.seasonality
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasSeasonality"] = [obj]
                

        if self.seasonalityGeneral:
            value_obj = self.seasonalityGeneral
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasSeasonalityGeneral"] = [obj]
                

        if self.seasonalityOriginal:
            value_obj = self.seasonalityOriginal
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasSeasonalityOriginal"] = [obj]
                

        if self.variable:
            value_obj = self.variable
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasVariable"] = [obj]
                

        if self.variableDetail:
            value_obj = self.variableDetail
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasVariableDetail"] = [obj]
                

        if self.variableGeneral:
            value_obj = self.variableGeneral
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasVariableGeneral"] = [obj]
                

        if self.variableGeneralDirection:
            value_obj = self.variableGeneralDirection
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasVariableGeneralDirection"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if self.basis:
            value_obj = self.basis
            obj = value_obj
            data["basis"] = obj

        if self.direction:
            value_obj = self.direction
            obj = value_obj
            data["direction"] = obj

        if self.local:
            value_obj = self.local
            obj = value_obj
            data["isLocal"] = obj

        if self.mathematicalRelation:
            value_obj = self.mathematicalRelation
            obj = value_obj
            data["mathematicalRelation"] = obj

        if self.notes:
            value_obj = self.notes
            obj = value_obj
            data["notes"] = obj

        if self.rank:
            value_obj = self.rank
            obj = value_obj
            data["rank"] = obj

        if self.scope:
            value_obj = self.scope
            obj = value_obj
            data["scope"] = obj

        if self.seasonality:
            value_obj = self.seasonality
            obj = value_obj.to_json()
            data["seasonality"] = obj

        if self.seasonalityGeneral:
            value_obj = self.seasonalityGeneral
            obj = value_obj.to_json()
            data["seasonalityGeneral"] = obj

        if self.seasonalityOriginal:
            value_obj = self.seasonalityOriginal
            obj = value_obj.to_json()
            data["seasonalityOriginal"] = obj

        if self.variable:
            value_obj = self.variable
            obj = value_obj.to_json()
            data["variable"] = obj

        if self.variableDetail:
            value_obj = self.variableDetail
            obj = value_obj
            data["variableDetail"] = obj

        if self.variableGeneral:
            value_obj = self.variableGeneral
            obj = value_obj
            data["variableGeneral"] = obj

        if self.variableGeneralDirection:
            value_obj = self.variableGeneralDirection
            obj = value_obj
            data["variableGeneralDirection"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'Interpretation':
        """Instantiate `Interpretation` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        Interpretation
            The populated `Interpretation` instance.
        """
        self = Interpretation()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "basis":
                    value = pvalue
                    obj = value
                    self.basis = obj
            elif key == "direction":
                    value = pvalue
                    obj = value
                    self.direction = obj
            elif key == "isLocal":
                    value = pvalue
                    obj = value
                    self.local = obj
            elif key == "mathematicalRelation":
                    value = pvalue
                    obj = value
                    self.mathematicalRelation = obj
            elif key == "notes":
                    value = pvalue
                    obj = value
                    self.notes = obj
            elif key == "rank":
                    value = pvalue
                    obj = value
                    self.rank = obj
            elif key == "scope":
                    value = pvalue
                    obj = value
                    self.scope = obj
            elif key == "seasonality":
                    value = pvalue
                    obj = InterpretationSeasonality.from_synonym(re.sub("^.*?#", "", value))
                    self.seasonality = obj
            elif key == "seasonalityGeneral":
                    value = pvalue
                    obj = InterpretationSeasonality.from_synonym(re.sub("^.*?#", "", value))
                    self.seasonalityGeneral = obj
            elif key == "seasonalityOriginal":
                    value = pvalue
                    obj = InterpretationSeasonality.from_synonym(re.sub("^.*?#", "", value))
                    self.seasonalityOriginal = obj
            elif key == "variable":
                    value = pvalue
                    obj = InterpretationVariable.from_synonym(re.sub("^.*?#", "", value))
                    self.variable = obj
            elif key == "variableDetail":
                    value = pvalue
                    obj = value
                    self.variableDetail = obj
            elif key == "variableGeneral":
                    value = pvalue
                    obj = value
                    self.variableGeneral = obj
            elif key == "variableGeneralDirection":
                    value = pvalue
                    obj = value
                    self.variableGeneralDirection = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getBasis(self) -> str:
        """Get basis.

        Returns
        -------
        str
            The current value of basis.
        """
        return self.basis

    def setBasis(self, basis:str):
        """Set basis.

        Parameters
        ----------
        basis : str
            The value to assign.
        """
        assert isinstance(basis, str), f"Error: '{basis}' is not of type str"
        self.basis = basis
    
    def getDirection(self) -> str:
        """Get direction.

        Returns
        -------
        str
            The current value of direction.
        """
        return self.direction

    def setDirection(self, direction:str):
        """Set direction.

        Parameters
        ----------
        direction : str
            The value to assign.
        """
        assert isinstance(direction, str), f"Error: '{direction}' is not of type str"
        self.direction = direction
    
    def getMathematicalRelation(self) -> str:
        """Get mathematicalRelation.

        Returns
        -------
        str
            The current value of mathematicalRelation.
        """
        return self.mathematicalRelation

    def setMathematicalRelation(self, mathematicalRelation:str):
        """Set mathematicalRelation.

        Parameters
        ----------
        mathematicalRelation : str
            The value to assign.
        """
        assert isinstance(mathematicalRelation, str), f"Error: '{mathematicalRelation}' is not of type str"
        self.mathematicalRelation = mathematicalRelation
    
    def getNotes(self) -> str:
        """Get notes.

        Returns
        -------
        str
            The current value of notes.
        """
        return self.notes

    def setNotes(self, notes:str):
        """Set notes.

        Parameters
        ----------
        notes : str
            The value to assign.
        """
        assert isinstance(notes, str), f"Error: '{notes}' is not of type str"
        self.notes = notes
    
    def getRank(self) -> str:
        """Get rank.

        Returns
        -------
        str
            The current value of rank.
        """
        return self.rank

    def setRank(self, rank:str):
        """Set rank.

        Parameters
        ----------
        rank : str
            The value to assign.
        """
        assert isinstance(rank, str), f"Error: '{rank}' is not of type str"
        self.rank = rank
    
    def getScope(self) -> str:
        """Get scope.

        Returns
        -------
        str
            The current value of scope.
        """
        return self.scope

    def setScope(self, scope:str):
        """Set scope.

        Parameters
        ----------
        scope : str
            The value to assign.
        """
        assert isinstance(scope, str), f"Error: '{scope}' is not of type str"
        self.scope = scope
    
    def getSeasonality(self) -> InterpretationSeasonality:
        """Get seasonality.

        Returns
        -------
        InterpretationSeasonality
            The current value of seasonality.
        """
        return self.seasonality

    def setSeasonality(self, seasonality:InterpretationSeasonality):
        """Set seasonality.

        Parameters
        ----------
        seasonality : InterpretationSeasonality
            The value to assign.
        """
        assert isinstance(seasonality, InterpretationSeasonality), f"Error: '{seasonality}' is not of type InterpretationSeasonality\nYou can create a new InterpretationSeasonality object from a string using the following syntax:\n- Fetch existing InterpretationSeasonality by synonym: InterpretationSeasonality.from_synonym(\"{seasonality}\")\n- Create a new custom InterpretationSeasonality: InterpretationSeasonality(\"{seasonality}\")"
        self.seasonality = seasonality
    
    def getSeasonalityGeneral(self) -> InterpretationSeasonality:
        """Get seasonalityGeneral.

        Returns
        -------
        InterpretationSeasonality
            The current value of seasonalityGeneral.
        """
        return self.seasonalityGeneral

    def setSeasonalityGeneral(self, seasonalityGeneral:InterpretationSeasonality):
        """Set seasonalityGeneral.

        Parameters
        ----------
        seasonalityGeneral : InterpretationSeasonality
            The value to assign.
        """
        assert isinstance(seasonalityGeneral, InterpretationSeasonality), f"Error: '{seasonalityGeneral}' is not of type InterpretationSeasonality\nYou can create a new InterpretationSeasonality object from a string using the following syntax:\n- Fetch existing InterpretationSeasonality by synonym: InterpretationSeasonality.from_synonym(\"{seasonalityGeneral}\")\n- Create a new custom InterpretationSeasonality: InterpretationSeasonality(\"{seasonalityGeneral}\")"
        self.seasonalityGeneral = seasonalityGeneral
    
    def getSeasonalityOriginal(self) -> InterpretationSeasonality:
        """Get seasonalityOriginal.

        Returns
        -------
        InterpretationSeasonality
            The current value of seasonalityOriginal.
        """
        return self.seasonalityOriginal

    def setSeasonalityOriginal(self, seasonalityOriginal:InterpretationSeasonality):
        """Set seasonalityOriginal.

        Parameters
        ----------
        seasonalityOriginal : InterpretationSeasonality
            The value to assign.
        """
        assert isinstance(seasonalityOriginal, InterpretationSeasonality), f"Error: '{seasonalityOriginal}' is not of type InterpretationSeasonality\nYou can create a new InterpretationSeasonality object from a string using the following syntax:\n- Fetch existing InterpretationSeasonality by synonym: InterpretationSeasonality.from_synonym(\"{seasonalityOriginal}\")\n- Create a new custom InterpretationSeasonality: InterpretationSeasonality(\"{seasonalityOriginal}\")"
        self.seasonalityOriginal = seasonalityOriginal
    
    def getVariable(self) -> InterpretationVariable:
        """Get variable.

        Returns
        -------
        InterpretationVariable
            The current value of variable.
        """
        return self.variable

    def setVariable(self, variable:InterpretationVariable):
        """Set variable.

        Parameters
        ----------
        variable : InterpretationVariable
            The value to assign.
        """
        assert isinstance(variable, InterpretationVariable), f"Error: '{variable}' is not of type InterpretationVariable\nYou can create a new InterpretationVariable object from a string using the following syntax:\n- Fetch existing InterpretationVariable by synonym: InterpretationVariable.from_synonym(\"{variable}\")\n- Create a new custom InterpretationVariable: InterpretationVariable(\"{variable}\")"
        self.variable = variable
    
    def getVariableDetail(self) -> str:
        """Get variableDetail.

        Returns
        -------
        str
            The current value of variableDetail.
        """
        return self.variableDetail

    def setVariableDetail(self, variableDetail:str):
        """Set variableDetail.

        Parameters
        ----------
        variableDetail : str
            The value to assign.
        """
        assert isinstance(variableDetail, str), f"Error: '{variableDetail}' is not of type str"
        self.variableDetail = variableDetail
    
    def getVariableGeneral(self) -> str:
        """Get variableGeneral.

        Returns
        -------
        str
            The current value of variableGeneral.
        """
        return self.variableGeneral

    def setVariableGeneral(self, variableGeneral:str):
        """Set variableGeneral.

        Parameters
        ----------
        variableGeneral : str
            The value to assign.
        """
        assert isinstance(variableGeneral, str), f"Error: '{variableGeneral}' is not of type str"
        self.variableGeneral = variableGeneral
    
    def getVariableGeneralDirection(self) -> str:
        """Get variableGeneralDirection.

        Returns
        -------
        str
            The current value of variableGeneralDirection.
        """
        return self.variableGeneralDirection

    def setVariableGeneralDirection(self, variableGeneralDirection:str):
        """Set variableGeneralDirection.

        Parameters
        ----------
        variableGeneralDirection : str
            The value to assign.
        """
        assert isinstance(variableGeneralDirection, str), f"Error: '{variableGeneralDirection}' is not of type str"
        self.variableGeneralDirection = variableGeneralDirection
    
    def isLocal(self) -> str:
        """Get local.

        Returns
        -------
        str
            The current value of local.
        """
        return self.local

    def setLocal(self, local:str):
        """Set local.

        Parameters
        ----------
        local : str
            The value to assign.
        """
        assert isinstance(local, str), f"Error: '{local}' is not of type str"
        self.local = local
    


from pylipd.lipd import LiPD
import json
from pylipd.lipd_series import LiPDSeries
from pylipd.utils.multi_processing import convert_to_rdf

####################
# TODO:
# - Edit local LiPD & update endpoint
####################


local_lipd_dir = "/Users/varun/git/LiPD/PyLiPD/data/lpd"
remote_lipd_endpoint = "https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse-dynamic"
#remote_lipd_endpoint = "https://localhost:7200/repositories/LiPDVerse-dynamic"

'''
lipd = LiPD()
# Convert LiPD files to RDF    
lipd.convert_lipd_dir_to_rdf(
    local_lipd_dir,
    local_lipd_dir+".nq", 
    parallel=False)

exit()
'''

'''
# Convert one LiPD file to RDF
convert_to_rdf(
    ("/Users/varun/Downloads/Pella.Lim.2016.lpd",
    "/Users/varun/Downloads/Pella.Lim.2016.nq")
)
exit()
'''


'''    
data = ['https://lipdverse.org/data/TjhHrDv0LQ4aazHolZkR/1_0_0//Ocn-WEqPacific.Stott.2007.lpd']
lipd.load(data)

ts_list = lipd.get_timeseries(lipd.get_all_dataset_names())
for dsname, tsos in ts_list.items():
    for tso in tsos:
        if 'paleoData_variableName' in tso:
            print(dsname+': '+tso['paleoData_variableName']+': '+tso['archiveType'])    

'''

# Load Datasets (from Local and Remote)
dsnames = ["MD98_2181.Stott.2007"]
remote_dsnames = ["Ocn-MadangLagoonPapuaNewGuinea.Kuhnert.2001"]

'''
L = LiPD()

#L.load(local_lipd_dir+"/"+"ODP1671017E.lpd")
#L.set_endpoint("https://linkedearth.graphdb.mint.isi.edu/repositories/LiPDVerse2")    
L.load([local_lipd_dir + "/" + dsname + ".lpd" for dsname in ["MD98_2181.Stott.2007","NAm-SmithersSkiArea.Schweingruber.1996", 
                        "NAm-CeaderBreaks.Briffa.1996", "ODP1671017E", 
                        "SPC14.Kahle.2021", "RC12-10.Poore.2003", 
                        "MD02-2553.Poore.2009", "AD9117.Guinta.2001", "SP10BEIN",
                        "SchellingsBog.Barron.2004", "Hidden.Tang.1999"]])
#bibtex = L.get_bibtex()
#print(bibtex)
#exit()

S = LiPDSeries()
S.load(L)

print(S)

exit()
'''

# Load from local
lipd = LiPD()

#lipd.load("/Users/varun/git/pylipd/pylipd/data/ODP846.Lawrence.2006.lpd")
lipd.load("/Users/varun/Downloads/Crystal.McCabe-Glynn.2013.lpd")
print(lipd.get_all_dataset_names())
ts, ts_df = lipd.get_timeseries(lipd.get_all_dataset_names(), to_dataframe=True)
print(ts_df.columns)
#print(ts_df[["dataSetName", "age", "archiveType", "paleoData_variableName", "paleoData_units"]])
ensemble_df = lipd.get_ensemble_tables()
df = lipd.get_timeseries_essentials(lipd.get_all_dataset_names()[0], mode='paleo')
print(df)
exit()

series = lipd.to_lipd_series()
print(series.get_all_variable_names())
print(series.get_timeseries_essentials())
exit()

# data_path = local_lipd_dir + '/Ocn-Palmyra.Nurhati.2011.lpd'
# lipd.load(data_path)

lipd.load("/Users/varun/Downloads/18.1.Franklin.Bolshoye.2024.lpd")
print(lipd.get_all_dataset_names())

#lipdfiles = [local_lipd_dir + "/" + dsname + ".lpd" for dsname in dsnames]
#print(lipdfiles)

#lipd.load(lipdfiles)
# lipd.load_from_dir("examples/data")

lipd.load_from_dir("/Users/varun/Downloads/example_sisal_lipds")
print(lipd.get_all_dataset_names())

lat = -77.08
lon = 38.91
radius = 50

(result, result_df) = lipd.query("""
PREFIX geo: <http://www.opengis.net/ont/geosparql#>
PREFIX geof: <http://www.opengis.net/def/function/geosparql/>
PREFIX wgs84: <http://www.w3.org/2003/01/geo/wgs84_pos#>

SELECT ?what
WHERE {
  ?ds wgs84:lat ?lat .
  ?ds wgs84:lon ?lon .
  
  FILTER(geof:sfWithin(?geometry,
     "POLYGON((-77.089005 38.913574,-77.029953 38.913574,-77.029953 38.886321,-77.089005 38.886321,-77.089005 38.913574))"^^geo:wktLiteral))
}
""")

#print(lipd.get_all_dataset_ids())
ens_df = lipd.get_ensemble_tables(
    ensembleVarName="age",
    ensembleDepthVarName="depth"
)
print(ens_df)


#lipd.load(["/Users/varun/Downloads/Arc-LakeNataujärvi.Ojala.2005.lpd"])
#print(lipd.get_all_dataset_names())

ts_list = lipd.get_timeseries(lipd.get_all_dataset_names())
for dsname, tsos in ts_list.items():
    for tso in tsos:
        if 'paleoData_variableName' in tso:
            print(dsname+': '+str(tso['paleoData_variableName'])+': '+tso['archiveType'])


# Fetch LiPD data from remote RDF Graph
#lipd.set_endpoint(remote_lipd_endpoint)
#lipd.load_remote_datasets(remote_dsnames)

# Convert to TSO object (as before)
'''
ts_list_remote = lipd.get_timeseries(lipd.get_all_dataset_names())
for dsname, tsos in ts_list_remote.items():
    for tso in tsos:
        print(dsname+': '+str(tso['paleoData_variableName'])+': '+tso['archiveType'])
'''

print("Popping ...")
poplipd = lipd.pop(lipd.get_all_dataset_names()[0])
print(lipd.get_all_dataset_names())
print(poplipd.get_all_dataset_names())

print("Creating separate lipd file for popped")
dsname = poplipd.get_all_dataset_names()[0]
lipdfile = f"{dsname}.lpd"
poplipd.create_lipd(dsname, lipdfile)

print("Loading .. created lipd file: "+lipdfile)
lpd2 = LiPD()
lpd2.load([lipdfile])
ts_list_remote = lpd2.get_timeseries(lpd2.get_all_dataset_names())
for dsname, tsos in ts_list_remote.items():
    for tso in tsos:
        print(dsname+': '+str(tso['paleoData_variableName'])+': '+tso['archiveType'])

ens_df2 = lpd2.get_ensemble_tables(
    ensembleVarName="age",
    ensembleDepthVarName="depth"
)
print(ens_df2)

print("After popping..")
print(lipd.get_all_dataset_names())
ts_list_remote = lipd.get_timeseries(lipd.get_all_dataset_names())
for dsname, tsos in ts_list_remote.items():
    for tso in tsos:
        print(dsname+': '+str(tso['paleoData_variableName'])+': '+tso['archiveType'])

print("Popped..")
print(poplipd.get_all_dataset_names())

print("Merging back..")
mergelipd = lipd.merge(poplipd)
print(mergelipd.get_all_dataset_names())

ts_list_remote = mergelipd.get_timeseries(mergelipd.get_all_dataset_names())
for dsname, tsos in ts_list_remote.items():
    for tso in tsos:
        print(dsname+': '+str(tso['paleoData_variableName'])+': '+tso['archiveType'])

'''
print(lipd.get_all_dataset_names())    
datasets = lipd.get_datasets(dsnames=dsnames)
print("Fetched..")
for ds in datasets:
    print(ds['id'])
    #print(json.dumps(ds, indent=3))
'''

# Usage
# - Just look at https://pyleoclim-util.readthedocs.io/en/master/core/api.html#lipdseries-pyleoclim-lipdseries
# - Implementing Series, MultipleSeries, EnsemebleSeries, Lipd, LipdSeries
# - Pyleoclim
#     - Given: Variable name, optional Dataset IDs
#       - Return time series data for the  matching variables (optional fuzzy match)
#     - Given: Proxy, optional Dataset IDs
#       - Return time series data for the  variables with matching proxies (optional fuzzy match)

#   - Time Series Data consists of:
#       - Variable value, Variable name, Variable unit, Time axis name (age/year), Time axis unit, Time axis values, Lat, Long, Dataset ID

#   - Lat, Long, Elevation
#   - Get the ensemble table given a dataset ID & age Year
# - Own metadata



##############################
# Auto-generated. Do not Edit
##############################

import re
import pandas as pd
import json
from pylipd.classes.variable import Variable
from pylipd.utils import uniqid
from pylipd.classes.variable import Variable

class DataTable:
    """Auto-generated LinkedEarth class representing `DataTable`."""
    def __init__(self):
        """Initialize a new DataTable instance."""
        
        self.fileName: str = None
        self.missingValue: str = None
        self.variables: list[Variable] = []
        self.misc = {}
        self.ontns = "http://linked.earth/ontology#"
        self.ns = "http://linked.earth/lipd"
        self.type = "http://linked.earth/ontology#DataTable"
        self.id = self.ns + "/" + uniqid("DataTable.")

    @staticmethod
    def from_data(id, data) -> 'DataTable':
        """Instantiate `DataTable` from an ontology-style data graph.

        Parameters
        ----------
        id : str
            The node identifier for this object.
        data : dict
            Dictionary mapping node ids to their predicate lists.

        Returns
        -------
        DataTable
            The populated `DataTable` instance.
        """
        self = DataTable()
        self.id = id
        mydata = data[id]
        for key in mydata:
            value = mydata[key]
            obj = None
            if key == "type":
                for val in value:
                    self.type = val["@id"]
        
            elif key == "hasFileName":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.fileName = obj
        
            elif key == "hasMissingValue":
                for val in value:
                    if "@value" in val:
                        obj = val["@value"]                        
                    self.missingValue = obj
        
            elif key == "hasVariable":
                for val in value:
                    if "@id" in val:
                        obj = Variable.from_data(val["@id"], data)
                    else:
                        obj = val["@value"]
            
                    self.variables.append(obj)
            else:
                for val in value:
                    obj = None
                    if "@id" in val:
                        obj = data[val["@id"]]
                    elif "@value" in val:
                        obj = val["@value"]
                    self.set_non_standard_property(key, obj)
            
        return self

    def to_data(self, data={}):
        """Serialize the object into a JSON-LD compatible dictionary.

        Parameters
        ----------
        data : dict, optional
            Existing data dictionary to extend.

        Returns
        -------
        dict
            The updated data dictionary.
        """
        data[self.id] = {}
        data[self.id]["type"] = [
            {
                "@id": self.type,
                "@type": "uri"
            }
        ]

        if len(self.variables):
            data[self.id]["hasVariable"] = []
        for value_obj in self.variables:
            if type(value_obj) is str:
                obj = {
                    "@value": value_obj,
                    "@type": "literal",
                    "@datatype": "http://www.w3.org/2001/XMLSchema#string"
                }
            else:
                obj = {
                    "@id": value_obj.id,
                    "@type": "uri"
                }
                data = value_obj.to_data(data)
            data[self.id]["hasVariable"].append(obj)

        if self.fileName:
            value_obj = self.fileName
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasFileName"] = [obj]
                

        if self.missingValue:
            value_obj = self.missingValue
            obj = {
                "@value": value_obj,
                "@type": "literal",
                "@datatype": "http://www.w3.org/2001/XMLSchema#string"
            }
            data[self.id]["hasMissingValue"] = [obj]
                
        for key in self.misc:
            value = self.misc[key]
            data[self.id][key] = []
            ptype = None
            tp = type(value).__name__
            if tp == "int":
                ptype = "http://www.w3.org/2001/XMLSchema#integer"
            elif tp == "float" or tp == "double":
                ptype = "http://www.w3.org/2001/XMLSchema#float"
            elif tp == "str":
                if re.match(r"\d{4}-\d{2}-\d{2}( |T)\d{2}:\d{2}:\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#datetime"   
                elif re.match(r"\d{4}-\d{2}-\d{2}", value):
                    ptype = "http://www.w3.org/2001/XMLSchema#date"
                else:
                    ptype = "http://www.w3.org/2001/XMLSchema#string"
            elif tp == "bool":
                ptype = "http://www.w3.org/2001/XMLSchema#boolean"

            data[self.id][key].append({
                "@value": value,
                "@type": "literal",
                "@datatype": ptype
            })
        
        return data

    def to_json(self):
        """Return a lightweight JSON representation (used by LiPD).

        Returns
        -------
        dict
            A dictionary representation of this object.
        """
        data = {
            "@id": self.id
        }

        if len(self.variables):
            data["columns"] = []
        for value_obj in self.variables:
            obj = value_obj.to_json()
            data["columns"].append(obj)

        if self.fileName:
            value_obj = self.fileName
            obj = value_obj
            data["filename"] = obj

        if self.missingValue:
            value_obj = self.missingValue
            obj = value_obj
            data["missingValue"] = obj

        for key in self.misc:
            value = self.misc[key]
            data[key] = value
                   
        return data

    @staticmethod
    def from_json(data) -> 'DataTable':
        """Instantiate `DataTable` from its lightweight JSON representation.

        Parameters
        ----------
        data : dict
            The JSON dictionary to parse.

        Returns
        -------
        DataTable
            The populated `DataTable` instance.
        """
        self = DataTable()
        for key in data:
            pvalue = data[key]
            if key == "@id":
                self.id = pvalue
            elif key == "columns":
                for value in pvalue:
                    obj = Variable.from_json(value)
                    self.variables.append(obj)
            elif key == "filename":
                    value = pvalue
                    obj = value
                    self.fileName = obj
            elif key == "missingValue":
                    value = pvalue
                    obj = value
                    self.missingValue = obj
            else:
                self.set_non_standard_property(key, pvalue)
                   
        return self

    def set_non_standard_property(self, key, value):
        """Store a predicate that is not defined in the ontology schema.

        This is useful for forward-compatibility with new properties that are
        not yet part of the official schema.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The property value.
        """
        if key not in self.misc:
            self.misc[key] = value
    
    def get_non_standard_property(self, key):
        """Return a single non-standard property by key.

        Parameters
        ----------
        key : str
            The property name.

        Returns
        -------
        any
            The property value.
        """
        return self.misc[key]
                
    def get_all_non_standard_properties(self):
        """Return the dictionary of all non-standard properties.

        Returns
        -------
        dict
            Dictionary of all non-standard properties.
        """
        return self.misc

    def add_non_standard_property(self, key, value):
        """Append a value to a list-valued non-standard property.

        Parameters
        ----------
        key : str
            The property name.
        value : any
            The value to append.
        """
        if key not in self.misc:
            self.misc[key] = []
        self.misc[key].append(value)
        
    def getFileName(self) -> str:
        """Get fileName.

        Returns
        -------
        str
            The current value of fileName.
        """
        return self.fileName

    def setFileName(self, fileName:str):
        """Set fileName.

        Parameters
        ----------
        fileName : str
            The value to assign.
        """
        assert isinstance(fileName, str), f"Error: '{fileName}' is not of type str"
        self.fileName = fileName
    
    def getMissingValue(self) -> str:
        """Get missingValue.

        Returns
        -------
        str
            The current value of missingValue.
        """
        return self.missingValue

    def setMissingValue(self, missingValue:str):
        """Set missingValue.

        Parameters
        ----------
        missingValue : str
            The value to assign.
        """
        assert isinstance(missingValue, str), f"Error: '{missingValue}' is not of type str"
        self.missingValue = missingValue
    
    def getVariables(self) -> list[Variable]:
        """Get variables list.

        Returns
        -------
        list[Variable]
            A list of Variable objects.
        """
        return self.variables

    def setVariables(self, variables:list[Variable]):
        """Set the variables list.

        Parameters
        ----------
        variables : list[Variable]
            The list to assign.
        """
        assert isinstance(variables, list), "Error: variables is not a list"
        assert all(isinstance(x, Variable) for x in variables), f"Error: '{variables}' is not of type Variable"
        self.variables = variables

    def addVariable(self, variables:Variable):
        """Add a value to the variables list.

        Parameters
        ----------
        variables : Variable
            The value to append.
        """
        assert isinstance(variables, Variable), f"Error: '{variables}' is not of type Variable"
        self.variables.append(variables)
        
    # Special Functions manually added for DataTable class
    def getDataFrame(self, use_standard_names=False) -> pd.DataFrame:
        cols = []
        for v in self.variables:
            colname = v.getName()
            if use_standard_names and v.getStandardVariable() is not None:
                colname = v.getStandardVariable().getLabel()
            cols.append(colname)
        
        df = pd.DataFrame(columns=cols)
        for v in self.variables:
            colname = v.getName()
            if use_standard_names and v.getStandardVariable() is not None:
                colname = v.getStandardVariable().getLabel()
            df[colname] = json.loads(v.getValues())
        
        # Create metadata as a dictionary and add to dataframe attr
        df.attrs = {}
        for v in self.variables:
            colname = v.getName()
            if use_standard_names and v.getStandardVariable() is not None:
                colname = v.getStandardVariable().getLabel()
            df.attrs[colname] = v.to_json()
            del df.attrs[colname]["hasValues"]
    
        return df

    def setDataFrame(self, df: pd.DataFrame):
        # Create new set of variable objects using the metadata
        self.variables = []
        for colname in df.attrs:
            v = Variable.from_json(df.attrs[colname])
            v.setValues(json.dumps(df[colname].to_list()))
            self.addVariable(v)



from tqdm import tqdm
from ..globals.queries import QUERY_ALL_VARIABLES_GRAPH
from .lipd_to_rdf import LipdToRDF
import multiprocessing as mp

def convert_to_rdf(files):
    (lipdfile, rdffile, standardize, add_labels) = files
    converter = LipdToRDF(standardize, add_labels)    
    """Worker that converts one lipdfile to an rdffile"""
    try:
        converter.convert(lipdfile)
        converter.serialize(rdffile)
    except Exception as e:
        print(f"ERROR: Could not convert LiPD file {lipdfile} to RDF")            
        raise e


def multi_convert_to_rdf(filemap, parallel=True, standardize=True, add_labels=True):
    if parallel:
        """Create a pool to convert all lipdfiles to rdffiles"""
        args = [(lipdfile, rdffile, standardize, add_labels) for lipdfile, rdffile in filemap.items()]
        pool = mp.Pool(mp.cpu_count())
        for file in tqdm(pool.imap_unordered(convert_to_rdf, args, chunksize=1), total=len(args)):
            pass
        pool.close()
    else:
        for lipdfile, rdffile in filemap.items():
            convert_to_rdf((lipdfile, rdffile, standardize, add_labels))


def convert_lipd_to_graph(arg):
    (lipdfile, standardize, add_labels) = arg
    """Worker that converts one lipdfile to an RDF graph"""
    try:
        converter = LipdToRDF(standardize, add_labels)
        converter.convert(lipdfile)
        return converter.graph
    except Exception as e:
        print(f"ERROR: Could not convert LiPD file {lipdfile} to RDF")            
        raise e


def multi_load_lipd(graph, lipdfiles, parallel=True, standardize=True, add_labels=True):
    """Load all lipdfiles to the RDF graph"""
    args = [(file, standardize, add_labels) for file in lipdfiles]
    if parallel:            
        with mp.Pool(mp.cpu_count()) as pool:
            for subgraph in tqdm(pool.imap_unordered(convert_lipd_to_graph, args, chunksize=1), total=len(lipdfiles)):
                graph.addN(subgraph.quads())
                del subgraph
            pool.close()
            pool.join()

    else:
        for i in tqdm(range(0, len(args))):
            arg = args[i]
            subgraph = convert_lipd_to_graph(arg)
            graph.addN(subgraph.quads())
            del subgraph
    return graph

    
def extract_variables_graph(arg):
    ctxid, subgraph = arg
    query = QUERY_ALL_VARIABLES_GRAPH
    subgraph.update(query)
    subgraph.remove((None, None, None, ctxid))
    return (ctxid, subgraph)


def multi_load_lipd_series(graph, single_dataset_lipds, parallel=True):
    """Load all lipd variables to the RDF graph"""

    args = []
    for ctxid, lipd in single_dataset_lipds.items():
        args.append((ctxid, lipd.graph))

    print("- Extracting variable subgraphs")
    if parallel:
        with mp.Pool(mp.cpu_count()) as pool:
            for (ctxid, subgraph) in tqdm(pool.imap_unordered(extract_variables_graph, args, chunksize=1), total=len(args)):
                # TODO: Add triples to store the dsname & tablename that the variable belongs to
                graph.addN(subgraph.quads())
            pool.close()
            pool.join()

    else:
        for i in tqdm(range(0, len(args))):
            ctxid, subgraph = args[i]
            ctxid, subgraph = extract_variables_graph(args[i])
            graph.addN(subgraph.quads())
    return graph

